#include "World.h"
#include "SubGenerator.h"

#include "Timer.h"

World* World::theWorld = 0;

#define PHYSICAL_HEIGHT_FACTOR	(10.0f);

World::World(void)
{
	m_currentHeight = 0;

	m_visualmode = VISUAL_TOPDOWN;
	m_rendermode = RENDER_SOLID;

	assert(theWorld == 0);
	theWorld = this;
	ClearDebugObjects();
}

World::~World(void)
{
	
}

void World::ResetCameras()
{
	camera.Reset();
	character.Reset();
}

VisualMode World::GetVisualMode()
{
	return m_visualmode;
}

void World::SetVisualMode(VisualMode selected)
{
	m_visualmode = selected;
}

void World::CycleVisualMode()
{
	m_visualmode = (VisualMode) ((m_visualmode + 1) % 2);
}

RenderMode World::GetRenderMode()
{
	return m_rendermode;
}

void World::SetRenderMode(RenderMode selected)
{
	m_rendermode = selected;
}

void World::CreateView()
{
	switch(m_visualmode)
	{
	case VISUAL_TOPDOWN:
		camera.CreateCamera();
		break;
	case VISUAL_CHARACTER:
		character.CreateCamera();
		break;
	}

	switch(m_rendermode)
	{
	case RENDER_SOLID:
		glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
		break;
	case RENDER_WIREFRAME:
		glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
		break;
	}
}

void World::GetGridContainingPoint(const Vector3d &point, float &x, float &y)
{
	GetGridContainingPoint(point.x, point.y, x, y);
}

void World::GetGridContainingPoint(float xpos, float ypos, float &x, float &y)
{
	x = GRID_SIZE * floor((xpos + GRID_SIZE/2) / GRID_SIZE);
	y = GRID_SIZE * floor((ypos + GRID_SIZE/2) / GRID_SIZE);
}

void World::GetCenterPosition(float &x, float &y)
{
	switch(m_visualmode)
	{
	case VISUAL_TOPDOWN:
		{
			x = camera.Lookat.x;
			y = camera.Lookat.y;
			return;
		}
		break;
	case VISUAL_CHARACTER:
	default:
		{
			x = character.Position.x;
			y = character.Position.y;
			return;
		}
		break;
	}
}

void World::GetWindowCenter(float &x, float &y)
{
	float xpos, ypos;
	GetCenterPosition(xpos, ypos);

	GetGridContainingPoint(xpos, ypos, x, y);
}

void World::PanMouse(float dx, float dy)
{
	switch(m_visualmode)
	{
	case VISUAL_TOPDOWN:
		{
			dx *= 10;
			dy *= 10;
			Vector3d thevector = ( camera.Lookat - camera.Position).Normalize();
			if(dx != 0)
			{
				camera.Lookat.x += thevector.y * dx;
				camera.Lookat.y -= thevector.x * dx;

				camera.Position.x += thevector.y * dx;
				camera.Position.y -= thevector.x * dx;
			}
			if(dy != 0)
			{
				camera.Lookat.x -= thevector.x * dy;
				camera.Lookat.y -= thevector.y * dy;

				camera.Position.x -= thevector.x * dy;
				camera.Position.y -= thevector.y * dy;
			}
		}
		break;
	case VISUAL_CHARACTER:
		{
			dx *= 4;
			dy *= 1;

			character.ChangeLook(dx, dy);
		}
		break;
	}

}

void World::ZoomMouse(float dx, float dy)
{
	switch(m_visualmode)
	{
	case VISUAL_TOPDOWN:
		{
			Vector3d outdirection = (camera.Position - camera.Lookat);
			float currentmag = outdirection.Magnitude();
			float newmag = currentmag + dy*10;
			newmag = MAX(newmag, MINIMUM_MAGNITUDE);

			camera.Position = (outdirection.Normalize() * newmag) + camera.Lookat;
		}
		break;
	case VISUAL_CHARACTER:
		{
			//nothing for now, maybe
			break;
		}
	}
}

void World::BuildLighting()
{
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHT1);

	//setting intensity
	float topintensity = 0.5f;
	GLfloat paramsIntensity[] = {topintensity, topintensity, topintensity, 1.0};
	glLightfv(GL_LIGHT1, GL_DIFFUSE, paramsIntensity);

	//setting position
	GLfloat paramsPosition[] = {camera.Position.x, camera.Position.y, camera.Position.z, 1};
	glLightfv(GL_LIGHT0, GL_POSITION, paramsPosition);

	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial ( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE ) ;

	if(m_visualmode == VISUAL_CHARACTER)
	{
		//add another light
		glEnable(GL_LIGHT2);

		//setting intensity
		float charintensity = 0.3f;
		GLfloat paramsIntensity2[] = {charintensity, charintensity, charintensity, 1.0};
		glLightfv(GL_LIGHT2, GL_DIFFUSE, paramsIntensity2);

		GLfloat paramsPosition2[] = {character.Position.x, character.Position.y, character.Position.z * 3, 1.0f};
		glLightfv(GL_LIGHT2, GL_POSITION, paramsPosition2);
	}
}

bool World::DoFrame(int timeElapsed, int *keysDown)
{
	bool didChange = false;

	if(m_visualmode == VISUAL_CHARACTER)
	{
		//check keyboard input
		if(keysDown['w'])
		{
			//forward
			character.Go(timeElapsed);
			didChange = true;
		}
		else if(keysDown['s'])
		{
			//backward
			character.Go(-timeElapsed);
			didChange = true;
		}

		//separately
		if(keysDown['a'])
		{
			//left
			character.Strafe(-timeElapsed);
			didChange = true;
		}
		else if(keysDown['d'])
		{
			//right
			character.Strafe(timeElapsed);
			didChange = true;
		}

		if(didChange)
		{
			//make topdown camera follow the character.
			Vector3d diff = character.Position - camera.Lookat;
			camera.Lookat = camera.Lookat + diff;
			camera.Position = camera.Position + diff;
		}
	}
	else
	{
		//top-down mode
#define VERTICAL_SPEED (0.001f)
		if(keysDown[' '])
		{
			//uplevel
			m_currentHeight += timeElapsed * VERTICAL_SPEED;
			didChange = true;
		}
		
		if(keysDown['c'])
		{
			//downlevel
			m_currentHeight -= timeElapsed * VERTICAL_SPEED;
			didChange = true;
		}
	}

	return didChange;
}

bool World::ConstrainToIndoors(Vector3d &originalPoint, Vector3d &delta)
{
	bool was_already_inside = m_architecture.ConstrainToIndoors(originalPoint, delta);

	return was_already_inside;
}

void World::SetLightingAndCamera()
{
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60, (GLfloat)m_width / (GLfloat)m_height, 0.01, 100.0);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	CreateView();

	BuildLighting();
}


void World::SetWindowSize(int newWindowSize)
{
	m_architecture.SetWindowSize(newWindowSize);
}

int World::GetWindowSize()
{
	return m_architecture.m_WindowSizeX;
}

void World::Render()
{
	//3D mode - show everything
	glClearColor (0.35f,0.35f,0.5f,1.0f);
	glClear (GL_COLOR_BUFFER_BIT);

	
	{
		AutoMatrixStack stack;
		m_architecture.Render();
	}

	float x, y;
	GetWindowCenter(x, y);


	if(m_visualmode == VISUAL_TOPDOWN)
	{
		float gridx, gridy;
		GetGridContainingPoint(character.Position.x, character.Position.y, gridx, gridy);
		float h = m_architecture.GetEntityHeight(gridx, gridy);
		character.RenderDirectionIndicator(h + m_currentHeight);
	}

#ifdef _DEBUG
	RenderDebugObjects();
#endif

}

void World::Update(bool force)
{
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	ClearDebugObjects();

	{
		AutoMatrixStack stack;
		m_architecture.Update(force);
	}
}

string World::GetLookDirection()
{
	return character.GetLookDirection();
}

float World::GetCurrentHeight()
{
	return m_currentHeight;
}

float World::GetLevelPhysicalHeight()
{
	return GetLevelPhysicalHeight(m_currentHeight);
}

float World::GetLevelPhysicalHeight(float fractionalHeight)
{
	return fractionalHeight * PHYSICAL_HEIGHT_FACTOR;
}


void World::AddDebugRay(Ray newray)
{
#ifdef _DEBUG
	if(m_rayCountDebug < MAX_DEBUG_OBJECTS)
	{
		m_debugRays[m_rayCountDebug++] = newray;
	}
#endif
}

void World::AddDebugPoint(Vector3d newpoint)
{
#ifdef _DEBUG
	if(m_pointCountDebug < MAX_DEBUG_OBJECTS)
	{
		m_debugPoints[m_pointCountDebug++] = newpoint;
	}
#endif
}

void World::RenderDebugObjects()
{
#ifdef _DEBUG
	for(int i = 0; i < m_pointCountDebug; i++)
	{
		m_debugPoints[i].DebugRender();
	}

	for(int i = 0; i < m_rayCountDebug; i++)
	{
		m_debugRays[i].DebugRender();
	}
#endif
}

void World::ClearDebugObjects()
{
	m_rayCountDebug = m_pointCountDebug = 0;
}